@@ -7,6 +7,7 @@
 */
 
 #include <fuse.h>
+#include <fuse_opt.h>
 
 #ifndef FUSE_VERSION
 #define FUSE_VERSION (FUSE_MAJOR_VERSION * 10 + FUSE_MINOR_VERSION)
@@ -19,8 +20,9 @@ typedef int (*fuse_cache_dirfil_t) (fuse_cache_dirh_t h, const char *name,
 struct fuse_cache_operations {
     struct fuse_operations oper;
     int (*cache_getdir) (const char *, fuse_cache_dirh_t, fuse_cache_dirfil_t);
-    
 };
 
 struct fuse_operations *cache_init(struct fuse_cache_operations *oper);
-int cache_parse_options(int *argcp, char *argv[]);
+int cache_parse_options(struct fuse_args *args);
+void cache_add_attr(const char *path, const struct stat *stbuf);
+void cache_invalidate(const char *path);
